import io
from operator import eq
from unittest import TestCase

import networkx as nx

from network import NetworkService
from network_analyser import NetworkAnalyser
from pyrthomas.cytoscape import read_sif, write_sif


class NetworkTestCase(TestCase):
    def setUp(self):
        self.network = nx.DiGraph()
        self.network.add_edge('x', 'y', weight=1)
        self.network.add_edge('x', 'x', weight=2)
        self.network.add_edge('y', 'x', weight=-1)

        # self.network.add_edge('BTK', 'MAL', weight=1)
        # self.network.add_edge('MAL', 'NFKB', weight=1)
        # self.network.add_edge('NFKB', 'INCY', weight=1)
        # self.network.add_edge('INCY', 'NFKB', weight=2)
        # self.network.add_edge('INCY', 'SOCS1', weight=1)
        # self.network.add_edge('SOCS1', 'NFKB', weight=-1)
        # self.network.add_edge('SOCS1', 'MAL', weight=-1)

    def test_state_space(self):
        result = NetworkAnalyser.get_state_space(self.network)
        expected = [{'x': 0, 'y': 0}, {'x': 0, 'y': 1}, {'x': 1, 'y': 0}, {'x': 1, 'y': 1},
                    {'x': 2, 'y': 0},
                    {'x': 2, 'y': 1}]
        self.assertEqual(result, expected)

    def test_state_graph(self):
        parameters = {
            "x": [
                ([], 0,), (['x'], 2), (['y'], 2), (['x', 'y'], 2)
            ],
            "y": [
                ([], 0,), (['x'], 1)
            ]
        }
        graph = NetworkAnalyser.get_state_graph(self.network, parameters)
        result = list(graph.nodes)
        expected = ['0,0', '0,1', '1,0', '1,1', '2,0', '2,1']
        self.assertListEqual(result, expected)

    def test_sif_converter(self):
        with io.StringIO('') as f:
            write_sif(self.network, f)
            f.seek(0)
            G = read_sif(f)
        em = nx.algorithms.isomorphism.generic_edge_match('weight', 1,
                                                          lambda x, y: eq(str(x), str(y)))
        result = nx.is_isomorphic(self.network, G, edge_match=em)
        self.assertTrue(result)

    def test_network_service_sif_converter(self):
        service = NetworkService()
        service.network = self.network
        with io.StringIO('') as f:
            service.export_sif(f)
            imported_network = NetworkService()

            f.seek(0)

            imported_network.import_sif(f)

        em = nx.algorithms.isomorphism.generic_edge_match('weight', 1,
                                                          lambda x, y: eq(str(x), str(y)))
        result = nx.is_isomorphic(self.network, imported_network.network, edge_match=em)
        self.assertTrue(result)

    def test_all(self):
        graphs = NetworkAnalyser.get_possible_state_graphs(self.network)
        assert len(list(graphs)) == 324



import itertools
from typing import Tuple, Dict, List

from networkx import DiGraph, simple_cycles
from networkx.classes.reportviews import EdgeView, NodeView

from pyrthomas import utils


class NetworkAnalyser:

    @staticmethod
    def set_predecessor_combinations(network: DiGraph):
        for node in network.nodes:
            predecessors = frozenset(DiGraph.predecessors(network, node))
            combinations = frozenset(utils.all_subsets(predecessors))
            nx.set_node_attributes(network, {node: combinations}, 'predecessor_combinations')

    @staticmethod
    def get_state_space(graph: DiGraph) -> List[Dict[str, int]]:
        max_thresholds = [NetworkAnalyser.get_max_threshold(graph, node) for node in graph.nodes]
        state_space = tuple(itertools.product(*max_thresholds))
        return [dict(zip(graph.nodes, state)) for state in state_space]

    @staticmethod
    def get_max_threshold(graph: DiGraph, node: NodeView) -> Tuple[int]:
        edges = graph.edges(node, data=True)
        max_weighted_edge = max(edges, key=lambda x: utils.get_weight(x, absolute=True))
        max_threshold = utils.get_weight(max_weighted_edge, True)
        return tuple(range(max_threshold + 1))

    @staticmethod
    def get_state_graph(network: DiGraph, parameters: Dict[str, List[Tuple[List[str], int]]]) -> DiGraph:
        """
        Args:
            network: Interaction graph in networkx format
            parameters: Parameters for state graph representing the current state of the entities

        Returns:
            Returns state graph in teh networkx format
        """
        state_space = NetworkAnalyser.get_state_space(network)
        state_graph = DiGraph()
        state_space_nodes = [utils.create_node_from_dict(state) for state in state_space]
        state_graph.add_nodes_from(state_space_nodes)

        resources = NetworkAnalyser.calculate_resources(network, state_space)

        k_states = NetworkAnalyser.calculate_k(resources, parameters)

        next_states = NetworkAnalyser.calculate_next_states(k_states)

        edges = NetworkAnalyser.generate_edges(next_states)
        state_graph.add_edges_from(edges)
        state_graph.graph['parameters'] = parameters
        return state_graph

    @staticmethod
    def generate_edges(next_states):
        edges = list()
        for key, states in next_states.items():
            previous_value = dict(key)
            source = utils.create_node_from_dict(previous_value)
            for state in states:
                target = utils.create_node_from_dict(state)
                edges.append((source, target))
        return edges

    @staticmethod
    def calculate_next_states(k_states):
        for state_key, state in k_states.items():
            previous_entities = dict(state_key)
            new_val = list()
            temp_previous = previous_entities.copy()
            for entity_key, previous_entity in previous_entities.items():
                next_value = k_states[state_key][entity_key]

                if next_value > previous_entity:
                    previous_entities[entity_key] = previous_entity + 1
                elif next_value < previous_entity:
                    previous_entities[entity_key] = previous_entity - 1

                if not (previous_entities in new_val) and not previous_entities == temp_previous:
                    new_val.append(previous_entities.copy())
                    previous_entities[entity_key] = temp_previous[entity_key]
            k_states[state_key] = new_val
        return k_states

    @staticmethod
    def calculate_k(resources, parameters):
        for resource_key, entities in resources.items():
            for entity_key, entity in entities.items():
                matched_interaction = filter(lambda x: sorted(x[0]) == sorted(entity),
                                             parameters[entity_key])
                first_interaction = next(matched_interaction)
                entities[entity_key] = first_interaction[1]
        return resources

    @staticmethod
    def get_cycles(network: DiGraph):
        return simple_cycles(network)

    @staticmethod
    def get_deadlock_states(network: DiGraph):
        out_degree_iter = network.out_degree(network.nodes)
        return [node for node, out_degree in out_degree_iter if out_degree == 0]

    @staticmethod
    def calculate_resources(network: DiGraph, state_space: List[Dict[str, int]]):
        resources = dict()
        for state in state_space:
            node = dict()
            for key in state:
                entity_resources = list()
                in_edges = network.in_edges(key, data=True)
                for edge in in_edges:
                    is_resource = NetworkAnalyser.is_resource_of_state(state, edge)
                    if is_resource:
                        entity_resources.append(edge[0])
                entity_resources.sort()
                node[key] = entity_resources
            state_key = tuple(state.items())
            resources[state_key] = node
        return resources

    @staticmethod
    def get_required_parameters(network: DiGraph):
        nodes = network.nodes
        parameters = dict()
        for node in nodes:
            predecessors = frozenset(DiGraph.predecessors(network, node))
            required_interactions = [(interaction, None) for interaction in
                                     utils.all_subsets(predecessors)]
            parameters[node] = required_interactions

        return parameters

    @staticmethod
    def is_resource_of_state(state: dict, edge: EdgeView):
        weight = utils.get_weight(edge)
        is_positive = weight >= 0
        value = state[edge[0]]
        return (is_positive and value >= abs(weight)) or (not is_positive and value < abs(weight))

    @staticmethod
    def get_possible_parameters(network: DiGraph):
        combinations = dict()
        for node in network.nodes:
            combinations[node] = list()
            max_threshold = utils.get_max_weighted_edge_threashold(network, node)
            predecessors = frozenset(DiGraph.predecessors(network, node))
            required_interactions = list(utils.all_subsets(predecessors))
            for arrangement in itertools.product(range(0, max_threshold + 1), repeat=len(required_interactions)):
                arranged_combination = list(zip(required_interactions, arrangement))
                combinations[node].append(arranged_combination)
        return [dict(zip(combinations, v)) for v in itertools.product(*combinations.values())]

    @staticmethod
    def get_possible_state_graphs(network: DiGraph):
        all_params = NetworkAnalyser.get_possible_parameters(network)
        for param in all_params:
            yield NetworkAnalyser.get_state_graph(network, param)



import sys

from networkx.utils import open_file
import networkx as nx


@open_file(0, mode='r')
def read_sif(path, connector=None):
    """Read graph in cytoscape SIF format from path.

    Parameters
    ----------
    path : file or string
       File or filename to write.

    Returns
    -------
    G : NetworkX MultiGraph or MultiDiGraph."""

    lines = (line for line in path)
    G = nx.DiGraph()
    for i in lines:
        l = i.split()
        if len(l) == 3 and connector:
            if l[1] != connector:
                continue
        if len(l) == 3:  # interaction
            # take " out of the name
            if l[0][0] == '"' and l[0][-1] == '"':
                l[0] = l[0][1:-1]
            if l[2][0] == '"' and l[2][-1] == '"':
                l[2] = l[2][1:-1]
        else:  # node only
            if len(l) != 1: continue
            if l[0][0] == '"' and l[0][-1] == '"':
                l[0] = l[0][1:-1]
            G.add_node(l[0])
        if len(l) < 3: continue
        try:
            list(G.nodes).index(l[0])
        except:
            G.add_node(l[0])
        try:
            list(G.nodes).index(l[2])
        except:
            G.add_node(l[2])
        G.add_edge(l[0], l[2], weight=l[1])
    return G


@open_file(1, mode='w')
def write_sif(G: nx.DiGraph, path, connector=None):
    """Write NetworkX graph G to cytoscape sif format on path.

    Path can be a string or a file handle.
    """
    for edge in list(G.edges(data=True)):
        link = connector if connector is not None else edge[2]['weight']
        path.write("%s\t%s\t%s\n" % (edge[0],
                                     link,
                                     edge[1]))
    return


def importCyAttributes(obj, filename='-', attr=None):
    """Imports attributes from a Cytoscape's attributes file

    @param obj: network
    @param filename: attributes file
    """
    if filename == '-':
        arq = sys.stdin
    else:
        arq = open(filename, 'r')
    if attr == None:
        attr = arq.readline().split()[0]
        if attr == 'cluster' or attr == 'Cluster':
            attr = ('cluster', 'cluster')
    else:
        arq.next()  # read header even if it's not used
    for i in arq:
        l = i.split()
        # TODO: error if n == None
        n = obj.findNode(l[0])
        if n != None:
            obj.nodes[n].setAttribute(attr, ' '.join(l[2:]))
    if filename != '-':
        arq.close()


def exportCyAttributes(obj, filename='-', attrname=('cluster', 'cluster'), s=""):
    """Exports attributes to a Cytoscape's attributes file

    @param obj: network
    @param filename: file to export
    @param attrname: attribute to write as header of Cytoscape's
    attributes file"""
    if filename == '-':
        arq = sys.stdout
    else:
        arq = open(filename, 'w')
    if isinstance(attrname, tuple):
        arq.write("%s\n" % attrname[-1])
    else:
        arq.write("%s\n" % attrname)
    for i in range(0, obj.nnodes):
        arq.write("%s\t=\t%s%s\n" % (obj.nodes[i].getName(), s,
                                     obj.nodes[i].getAttribute(attrname)))
    if filename != '-':
        arq.close()



import time


def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            print('%r  %2.2f ms' % (method.__name__, (te - ts) * 1000))
        return result

    return timed



import itertools
from typing import FrozenSet, Dict

import networkx as nx
from networkx.classes.reportviews import EdgeView


def all_subsets(items: FrozenSet):
    return itertools.chain(
        *map(lambda x: itertools.combinations(items, x), range(0, len(items) + 1)))


def create_node_from_dict(item: Dict) -> str:
    return ','.join(map(str, item.values()))


def get_weight(edge: EdgeView, absolute=False) -> int:
    weight = edge[2]['weight']
    return abs(weight) if absolute else weight


def get_max_weighted_edge_threashold(network: nx.DiGraph, node: str) -> int:
    edges_from_node = network.edges(node, data=True)
    max_weighted_edge = max(edges_from_node, key=lambda x: get_weight(x, True))
    return get_weight(max_weighted_edge, True)



import networkx as nx

from pyrthomas.cytoscape import read_sif, write_sif


class NetworkService:
    network: nx.DiGraph

    def __init__(self) -> None:
        self.network = nx.DiGraph()

    def add_node(self, node):
        self.network.add_node(node.id, min=node.min, max=node.max)

    def add_edge(self, edge):
        self.network.add_edge(edge.source, edge.target, weight=edge.weight)

    def get_nodes(self):
        return list(self.network.nodes)

    def clear(self):
        self.network.clear()

    def get_edges(self):
        return list(self.network.edges)

    def import_graphml(self, file):
        self.network = nx.read_graphml(file)

    def import_dot(self, file):
        self.network = nx.drawing.nx_pydot.read_dot(file)

    def import_sif(self, file):
        self.network = read_sif(file)

    def export_graphml(self, file):
        nx.write_graphml(self.network, file, prettyprint=True)

    def export_dot(self, file):
        nx.drawing.nx_pydot.write_dot(self.network, file)

    def export_sif(self, file):
        write_sif(self.network, file)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyrthomas']

package_data = \
{'': ['*'], 'pyrthomas': ['UNKNOWN.egg-info/*']}

install_requires = \
['networkx>=2.0']

setup_kwargs = {
    'name': 'pyrthomas',
    'version': '0.0.9',
    'description': "Python implementation for René Thomas's Framework",
    'long_description': "Python implementation for Rene` Thomas's Framework\n\nFor installation execute\n\n`pip install pyrthomas`\n\nTo learn how to use this package please check examples directory.\n\n",
    'author': 'Talha Junaid',
    'author_email': 'talhajunaidd@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/talhajunaidd/pyrthomas',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



class Node(object):
    min = 0
    max = 0
    id = ''

    def __init__(self, node_id, node_min=0, node_max=0):
        self.id = node_id
        self.min = node_min
        self.max = node_max


class Edge(object):
    source = ''
    target = ''
    weight = 0

    def __init__(self, source, target, weight):
        self.target = target
        self.source = source
        self.weight = weight
